import os
import re

import six
from six.moves.urllib import parse

from jirafs import exceptions, utils
from jirafs.plugin import CommandPlugin
from jirafs.ticketfolder import TicketFolder


class Command(CommandPlugin):
    """ Clone a new ticketfolder for the specified ticket URL"""
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'
    AUTOMATICALLY_INSTANTIATE_FOLDER = False

    def handle(self, args, jira, path, **kwargs):
        ticket_url = args.ticket_url[0]
        ticket_url_parts = parse.urlparse(ticket_url)
        if not ticket_url_parts.netloc:
            default_server = utils.get_default_jira_server()
            ticket_url = parse.urljoin(
                default_server,
                'browse/' + ticket_url + '/'
            )
        path = args.path[0] if args.path else None

        return self.clone(path, ticket_url, jira)

    def clone(self, path, ticket_url, jira):
        match = re.match('.*\/browse\/(\w+-\d+)\/?', ticket_url)
        if not match:
            raise exceptions.JirafsError(
                "\'%s\' is not a valid JIRA ticket URL." % (
                    ticket_url
                )
            )
        if not path:
            path = match.group(1)
        path = os.path.realpath(path)
        os.mkdir(path)
        folder = TicketFolder.initialize_ticket_folder(ticket_url, path, jira)

        utils.run_command_method_with_kwargs('pull', folder=folder)

        return folder

    def add_arguments(self, parser):
        parser.add_argument(
            'ticket_url',
            nargs=1,
            type=six.text_type
        )
        parser.add_argument(
            'path',
            nargs='*',
            type=six.text_type,
        )



from textwrap import dedent
from unittest import TestCase

from mock import patch

from jirafs.jirafieldmanager import LocalFileJiraFieldManager


class TestJiraFieldManager(TestCase):
    def setUp(self):
        with patch(
            'jirafs.jirafieldmanager.JiraFieldManager.load'
        ):
            self.jirafieldmanager = LocalFileJiraFieldManager(None, None)

    def test_decode(self):
        encoded_values = dedent("""
            * summary:
                This is a test summary
            * longer_message:
                This is a much
                longer message that happens
                to contain newlines.
        """)

        expected_result = {
            'summary': 'This is a test summary',
            'longer_message': (
                'This is a much\n'
                'longer message that happens\n'
                'to contain newlines.'
            )
        }

        actual_result = self.jirafieldmanager.get_fields_from_string(
            encoded_values
        )

        self.assertEqual(
            expected_result,
            actual_result
        )



import os
import shutil
import tempfile

import mock
from mock import patch

from jirafs.ticketfolder import TicketFolder
from jirafs.utils import run_command_method_with_kwargs

from .base import BaseTestCase


class TestPlugins(BaseTestCase):
    def setUp(self):
        self.arbitrary_ticket_number = 'ALPHA-123'
        self.root_folder = tempfile.mkdtemp()
        self.mock_jira = mock.MagicMock()
        self.mock_jira.issue.return_value = self.rehydrate_issue(
            'basic.issue.json'
        )
        self.mock_get_jira = lambda _, config=None: self.mock_jira

        with patch(
            'jirafs.ticketfolder.TicketFolder.get_remotely_changed'
        ) as get_remotely_changed:
            get_remotely_changed.return_value = []
            self.ticketfolder = run_command_method_with_kwargs(
                'clone',
                ticket_url='http://arbitrary.com/browse/ALPHA-123',
                jira=self.mock_get_jira,
                path=os.path.join(
                    self.root_folder,
                    self.arbitrary_ticket_number,
                )
            )

    def test_load_plugins(self):
        existing_plugins = {
            'alpha': mock.Mock(
                MAX_VERSION='10.0', MIN_VERSION='0.1'
            ),
            'beta': mock.Mock(
                MAX_VERSION='10.0', MIN_VERSION='0.1'
            ),
            'delta': mock.Mock(
                MAX_VERSION='10.0', MIN_VERSION='0.1'
            ),
        }

        with patch('jirafs.utils.get_installed_plugins') as gip:
            gip.return_value = existing_plugins
            with patch.object(self.ticketfolder, 'get_config') as config:
                config.return_value = mock.Mock(
                    has_section=mock.Mock(return_value=True),
                    items=mock.Mock(
                        return_value=[
                            ('alpha', 'on'),
                            ('delta', 'enabled'),
                            ('gamma', 'yes'),
                        ]
                    )
                )
                results = self.ticketfolder.load_plugins()

        self.assertEquals(2, len(results))
        self.assertTrue(existing_plugins['alpha'].called)
        self.assertFalse(existing_plugins['beta'].called)
        self.assertTrue(existing_plugins['delta'].called)

    def tearDown(self):
        shutil.rmtree(self.root_folder)



import pydoc

from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Print the log for this issue """
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, folder, **kwargs):
        return self.log(folder)

    def log(self, folder):
        results = folder.get_log()
        pydoc.pager(results)
        return results



import argparse
import logging
import os
import sys
import time

import six
from verlib import NormalizedVersion

from . import utils
from .exceptions import (
    GitCommandError,
    JirafsError,
    NotTicketFolderException
)


logger = logging.getLogger(__name__)


def main():
    if sys.version_info < (2, 7):
        raise RuntimeError(
            "Jirafs requires minimally version 2.7 of Python 2, or "
            "any version of Python 3.  Please upgrade your version of "
            "python before using Jirafs."
        )
    if utils.get_git_version() < NormalizedVersion('1.8'):
        raise RuntimeError(
            "Jirafs requires minimally version 1.8 of Git.  Please "
            "upgrade your version of git before using Jirafs."
        )

    commands = utils.get_installed_commands()

    parser = argparse.ArgumentParser(
        description='Edit Jira issues locally from your filesystem',
        add_help=False,
    )
    parser.add_argument(
        'command',
        nargs=1,
        type=six.text_type,
        choices=commands.keys()
    )
    args, extra = parser.parse_known_args()

    command_name = args.command[0]
    cmd_class = commands[command_name]

    started = time.time()
    logger.debug(
        'Command %s(%s) started',
        command_name,
        extra
    )
    jira = utils.lazy_get_jira()
    try:
        cmd_class.execute_command(
            extra, jira=jira, path=os.getcwd(), command_name=command_name
        )
    except GitCommandError as e:
        print(
            "Error (code: %s) while running git command." % (
                e.returncode
            )
        )
        print("")
        print("Command:")
        print("    %s" % e.command)
        print("")
        print("Output:")
        for line in e.output.decode('utf8').split('\n'):
            print("    %s" % line)
        print("")
        sys.exit(1)
    except NotTicketFolderException:
        if not getattr(cmd_class, 'TRY_SUBFOLDERS', False):
            print(
                "The command '%s' must be ran from within an issue folder." % (
                    command_name
                )
            )
            sys.exit(1)
        count_runs = 0
        for folder in os.listdir(os.getcwd()):
            try:
                cmd_class.execute_command(
                    extra,
                    jira=jira,
                    path=os.path.join(
                        os.getcwd(),
                        folder,
                    ),
                    command_name=command_name,
                )
                count_runs += 1
            except NotTicketFolderException:
                pass
        if count_runs == 0:
            print(
                "The command '%s' must be ran from within an issue folder "
                "or from within a folder containing issue folders." % (
                    command_name
                )
            )
            sys.exit(1)
    except JirafsError as e:
        print("Jirafs encountered an error processing your request: %s" % e)

    logger.debug(
        'Command %s(%s) finished in %s seconds',
        command_name,
        extra,
        (time.time() - started)
    )



import mock

from jirafs import utils

from .base import BaseTestCase


class TestStashLocalChanges(BaseTestCase):
    def test_stash_local_changes(self):
        repo = mock.Mock()
        repo.version = 10

        with utils.stash_local_changes(repo):
            pass

        self.assertEqual(
            repo.run_git_command.call_count,
            3,
        )
        repo.run_git_command.assert_has_calls([
            mock.call('stash', '--include-untracked', failure_ok=True),
            mock.call('stash', 'apply', failure_ok=True),
            mock.call('stash', 'drop', failure_ok=True),
        ])

    def test_stash_local_changes_exception(self):
        repo = mock.Mock()
        repo.version = 10

        with self.assertRaises(Exception):
            with utils.stash_local_changes(repo):
                raise Exception()

        self.assertEqual(
            repo.run_git_command.call_count,
            3,
        )
        repo.run_git_command.assert_has_calls([
            mock.call('stash', '--include-untracked', failure_ok=True),
            mock.call('stash', 'apply', failure_ok=True),
            mock.call('stash', 'drop', failure_ok=True),
        ])



import six

from jirafs import constants, exceptions, utils
from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """ Push locally-committed changes to JIRA """
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, folder, **kwargs):
        return self.push(folder)

    def push(self, folder):
        with utils.stash_local_changes(folder):
            status = folder.status()

            if not folder.is_up_to_date():
                raise exceptions.LocalCopyOutOfDate()

            file_meta = folder.get_remote_file_metadata(shadow=False)

            for filename in status['ready']['files']:
                upload = six.BytesIO(
                    folder.get_local_file_at_revision(
                        filename,
                        'HEAD',
                        binary=True
                    )
                )
                filename, upload = folder.execute_plugin_method_series(
                    'alter_file_upload',
                    args=((filename, upload, ), ),
                    single_response=True,
                )
                folder.log(
                    'Uploading file "%s"',
                    (filename, ),
                )
                # Delete the existing issue if there is one
                for attachment in folder.issue.fields.attachment:
                    if attachment.filename == filename:
                        attachment.delete()
                upload.seek(0)
                attachment = folder.jira.add_attachment(
                    folder.ticket_number,
                    upload,
                    filename=filename,
                )
                file_meta[filename] = attachment.created

            folder.set_remote_file_metadata(file_meta, shadow=False)

            comment = folder.get_new_comment(clear=True, ready=True)
            if comment:
                folder.log('Adding comment "%s"' % comment)
                folder.jira.add_comment(folder.ticket_number, comment)

            collected_updates = {}
            for field, diff_values in status['ready']['fields'].items():
                collected_updates[field] = diff_values[1]

            if collected_updates:
                folder.log(
                    'Updating fields "%s"',
                    (collected_updates, )
                )
                folder.issue.update(**collected_updates)

            # Commit local copy
            folder.run_git_command('reset', '--soft', failure_ok=True)
            folder.run_git_command(
                'add', '.jirafs/remote_files.json', failure_ok=True
            )
            folder.run_git_command(
                'add', constants.TICKET_NEW_COMMENT, failure_ok=True
            )
            folder.run_git_command(
                'commit', '-m', 'Pushed local changes', failure_ok=True
            )

            # Commit changes to remote copy, too, so we record remote
            # file metadata.
            folder.run_git_command('fetch', shadow=True)
            folder.run_git_command('merge', 'origin/master', shadow=True)
            folder.run_git_command('add', '-A', shadow=True)
            folder.run_git_command(
                'commit', '-m', 'Pulled remote changes',
                failure_ok=True, shadow=True
            )
            folder.run_git_command('push', 'origin', 'jira', shadow=True)
            pull_result = run_command_method_with_kwargs('pull', folder=folder)
            return pull_result[1]



import os
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

from jirafs import __version__ as version_string


requirements_path = os.path.join(
    os.path.dirname(__file__),
    'requirements.txt',
)
try:
    from pip.req import parse_requirements
    requirements = [
        str(req.req) for req in parse_requirements(requirements_path)
    ]
except ImportError:
    requirements = []
    with open(requirements_path, 'r') as in_:
        requirements = [
            req for req in in_.readlines()
            if not req.startswith('-')
            and not req.startswith('#')
        ]


class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox
        errno = tox.cmdline(self.test_args)
        sys.exit(errno)


setup(
    name='jirafs',
    version=version_string,
    url='https://github.com/coddingtonbear/jirafs',
    description=(
        'Edit Jira issues as text files locally.'
    ),
    author='Adam Coddington',
    author_email='me@adamcoddington.net',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
    install_requires=requirements,
    tests_require=['tox'],
    cmdclass = {'test': Tox},
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'jirafs = jirafs.cmdline:main'
        ],
        'jirafs_commands': [
            'fetch = jirafs.commands.fetch:Command',
            'pull = jirafs.commands.pull:Command',
            'merge = jirafs.commands.merge:Command',
            'commit = jirafs.commands.commit:Command',
            'push = jirafs.commands.push:Command',
            'git = jirafs.commands.git:Command',
            'log = jirafs.commands.log:Command',
            'debug = jirafs.commands.debug:Command',
            'plugins = jirafs.commands.plugins:Command',
            'status = jirafs.commands.status:Command',
            'clone = jirafs.commands.clone:Command',
            'open = jirafs.commands.open:Command',
            'diff = jirafs.commands.diff:Command',
            'config = jirafs.commands.config:Command',
            'transition = jirafs.commands.transition:Command',
        ]
    },
)



from collections import OrderedDict

from six.moves import input

from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """ Transition the current issue into a new state """
    MIN_VERSION = '1.0'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        state = self.get_state_from_string(folder, args.state)
        if state is None:
            state = self.get_state_from_user(folder)
        return self.transition(folder, state)

    def transition(self, folder, state_id):
        folder.jira.transition_issue(folder.issue, state_id)
        pull_result = run_command_method_with_kwargs('pull', folder=folder)
        return pull_result[1]

    def get_transition_dict(self, folder):
        if not hasattr(self, '_transition_dict'):
            value = folder.jira.transitions(folder.issue)
            self._transition_dict = OrderedDict(
                (v['id'], v) for v in value
            )
        return self._transition_dict

    def get_state_from_string(self, folder, value):
        options = self.get_transition_dict(folder)

        if value is None:
            return None

        # If it's a dictionary key, just go for it.
        if value in options:
            return value

        # But, if it isn't, this'll be a bit less efficient
        for k, v in options.items():
            if value.upper() == v['name'].upper():
                return k

        return None

    def get_state_from_user(self, folder):
        options = self.get_transition_dict(folder)
        response = None
        while(response is None):
            for option_id, option_data in options.items():
                print(
                    "%s: %s (%s)" % (
                        option_id,
                        option_data['name'],
                        option_data.get('to', {}).get('description', ''),
                    )
                )

            print('')
            response = self.get_state_from_string(
                folder,
                input("Please select a state from the above options: ")
            )

        return response

    def add_arguments(self, parser):
        parser.add_argument(
            'state',
            default=None,
            nargs='?'
        )



import os
import re

from jirafs import constants


class JiraFieldManager(dict):
    FIELD_MATCHER = re.compile(
        '^%s$' % (
            constants.TICKET_FILE_FIELD_TEMPLATE.replace(
                '{field_name}', '([\w_]+)'
            )
        )
    )

    def __init__(self):
        self._data = self.load()
        super(JiraFieldManager, self).__init__(self._data)

    def __sub__(self, other):
        differing = {}
        for k, v in other.items():
            if self.get(k) != v:
                differing[k] = (v, self.get(k), )

        return differing

    @classmethod
    def create(cls, folder, revision=None, path=None):
        if revision and path:
            raise TypeError(
                'You may specify a git revision or a local path; not both.'
            )

        if revision:
            return GitRevisionJiraFieldManager(folder, revision)
        else:
            return LocalFileJiraFieldManager(folder, path)

    def get_requested_per_ticket_fields(self):
        return constants.FILE_FIELDS

    def get_used_per_ticket_fields(self):
        raise NotImplementedError()

    def get_fields_from_string(self, string):
        """ Gets field data from an incoming string.

        Parses through the string using the following RST-derived
        pattern::

            0 | * Field
            1 |     VALUE
            2 |     MORE VALUE

        """
        data = {}
        field_name = ''
        value = ''
        if not string:
            return data
        lines = string.split('\n')
        for idx, line in enumerate(lines):
            if line.startswith('*'):
                if value:
                    data[field_name] = value.strip()
                    value = ''
                field_name = re.match('^\* (\w+):$', line).group(1)
            elif field_name:
                value = value + '\n' + line.strip()
        if value:
            data[field_name] = value.strip()

        return data

    def load(self):
        fields = self.get_fields_from_string(
            self.get_file_contents(constants.TICKET_DETAILS)
        )

        used_fields = set(self.get_used_per_ticket_fields())
        requested_fields = set(self.get_requested_per_ticket_fields())

        file_fields = {}
        for field_name in used_fields | requested_fields:
            try:
                field_path = constants.TICKET_FILE_FIELD_TEMPLATE.format(
                    field_name=field_name
                )
                file_fields[field_name] = self.get_file_contents(field_path)
            except (IOError, OSError):
                pass

        fields.update(file_fields)
        return fields


class LocalFileJiraFieldManager(JiraFieldManager):
    def __init__(self, folder, path):
        self.folder = folder
        self.path = path
        super(LocalFileJiraFieldManager, self).__init__()

    def get_file_contents(self, path):
        full_path = os.path.join(self.path, path)

        with open(self.folder.get_local_path(full_path), 'r') as _in:
            return _in.read().strip()

    def get_used_per_ticket_fields(self):
        fields = []
        for filename in os.listdir(self.path):
            full_path = os.path.join(self.path, filename)
            matched = self.FIELD_MATCHER.match(filename)
            if matched and os.path.isfile(full_path):
                field_name = matched.group(1)
                if not field_name in constants.FILE_FIELD_BLACKLIST:
                    fields.append(field_name)

        return fields

    def save(self):
        raise NotImplementedError()


class GitRevisionJiraFieldManager(JiraFieldManager):
    def __init__(self, folder, revision):
        self.folder = folder
        self.revision = revision
        super(GitRevisionJiraFieldManager, self).__init__()

    def get_file_contents(self, path):
        return self.folder.get_local_file_at_revision(
            path,
            self.revision
        )

    def get_used_per_ticket_fields(self):
        files = self.folder.run_git_command(
            'ls-tree',
            '--name-only',
            self.revision
        ).split()
        fields = []

        for filename in files:
            matched = self.FIELD_MATCHER.match(filename)
            if matched:
                field_name = matched.group(1)
                if not field_name in constants.FILE_FIELD_BLACKLIST:
                    fields.append(field_name)

        return fields
